/* IO Library $Id: io.c,v 1.1.1.1 2004/11/13 23:28:02 advanced Exp $ */

#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<unistd.h>
#include<string.h>
#include<assert.h>
#include<stdio.h>

#include<btree/io.h>
#include<utils.h>

IO* io_fopen(const char *filename)
{
	IO *hd;

	hd = (IO*)u_malloc(sizeof(IO));

	hd->filename = u_strdup(filename);

	hd->fd = open(filename,O_CREAT|O_EXCL|O_RDWR,00644);
	if(hd->fd > 0){
		iostat = IO_CREAT;
		return hd;
	}

	hd->fd = open(filename,O_RDWR);
	if(hd->fd > 0){
		iostat = IO_OK;
		return hd;
	}

	hd->fd = open(filename,O_RDONLY);
	if(hd->fd > 0){
		iostat = IO_RDONLY;
		return hd;
	}

	iostat = IO_NOFILE;
	free(hd);
	return (IO*)NULL;
}


IO* io_fopen_rdonly(const char *filename)
{
	IO *hd;

	hd = (IO*)u_malloc(sizeof(IO));

	hd->fd = open(filename,O_RDONLY);
	if(hd->fd > 0){
		iostat = IO_RDONLY;
		return (IO*)hd;
	}
	
	iostat = IO_NOFILE;
	free(hd);
	return (IO*)NULL;
}

void io_fclose(IO *hd)
{
	close(hd->fd);
	free(hd);
}

int io_fread(IO *hd,void *buffer,size_t size)
{
	unsigned int toproc_size=0;
	unsigned int proc_size=0;
	unsigned int proc_sum=0;

	if(size > BUFSIZ)
	{
		toproc_size = BUFSIZ;

		do{
			if(size < BUFSIZ)
				toproc_size = size;

			proc_size = read(hd->fd,buffer,toproc_size);
			buffer += proc_size;
			size -= proc_size;
			proc_sum += proc_size;
		}while(size>0);

		return proc_sum;
	}

	return read(hd->fd,buffer,size);
}


/* write() 시스템 콜이 큰 사이즈의 파일을 쓸 경우에 한번에 쓸수 있는
   최대 한도 만큼만 쓰고 종료하는데 반해 함수 한번으로 여러번 나누어서 
   쓰게끔 구현하였다.
   
   ※  BUFSIZ 의 크기 만큼씩 나누어 쓰는것이 가장 효율적인지 확신이 없음 */
int io_fwrite(IO *hd,void *buffer,size_t size)
{
	unsigned int toproc_size=0;
	unsigned int proc_size=0;
	unsigned int proc_sum=0;

	if(size > BUFSIZ)
	{
		toproc_size = BUFSIZ;

		do{
			if(size < BUFSIZ)
				toproc_size = size;

			proc_size = write(hd->fd,buffer,toproc_size);
			buffer+=proc_size;
			size-=proc_size;
			proc_sum+=proc_size;

		}while(size>0);

		return proc_sum;
	}

	return write(hd->fd,buffer,size);
}


off_t io_fseek(IO *hd,off_t offset)
{
#ifdef _DEBUG
	off_t off;

	off = lseek(hd->fd,offset,SEEK_SET);
	fprintf(stdout,"io_fseek: %d\n",(int)off);
	return off;
#endif
	return lseek(hd->fd,offset,SEEK_SET);
}


int io_fexist(const char *filename)
{
	int fd;

	fd = open(filename,O_CREAT|O_EXCL,00644);

	if(fd <= 0){
		return IO_EXIST;
	}
	return IO_NOFILE;
}
