/* $Id: relation.c,v 1.1.1.1 2004/11/13 23:28:02 advanced Exp $

 Relation And Tuple Implement Code 
 Author : Hyunsik Choi(advanced)
 Date : 2004.10.6

*/

#include<sys/types.h>
#include<sys/stat.h>
#include<sys/fcntl.h>
#include<unistd.h>
#include<string.h>
#include<stdio.h>

#include<btree/relation.h>
#include<btree/errno.h>
#include<utils.h>
#include<type.h>

const char TINY_FM_IDENTITY[2] = {0x76,0x69};
const uint16 TINY_FM_VERSION = 0x0001;

/* 저장하기 위한 파일 헤더(12byte) 
 
 * identity - 식별자 0x76 0x69 [t][i]nySQL
 * version - 버전 0x00 0x01 version 1 을 의미
 * off_idx - 파일에서 tree index 영역 시작 offset
 * off_data -  파일에서 data 영역 시작 offset */

#pragma pack(1)
/* relname.db 파일의 헤더로써 Relation , 즉 table의 관한 위와 같은 
   정보를 담고 있다 */
typedef struct RelationHeader_ {
	unsigned char identity[2]; /* 식별자 */
	uint16 version;
/*	offset off_idx; */
	offset off_data;
	uint32 rows;
} RelationHeader;
#pragma pack()

/* Relation 구조체의 생성자로써 table를 create 또는 open 시 모두
   먼저 수행되어야 한다 */
Relation* rel_new(const char *relname)
{
	Relation *rel;

	rel = (Relation*)u_malloc(sizeof(Relation));

	rel->relname = u_strdup(relname);
	rel->db_file = u_strcat(relname,FEXT_DB);
	rel->dic_file = u_strcat(relname,FEXT_DIC);
	
	rel->io = NULL;
	
	rel->rows = 0;
	rel->szrec = 0;

/*	rel->off_idx = 0; */
	rel->off_data = 0;

	rel->attrlist = (AttrList*)NULL;
	rel->count_of_attr = 0;

	return rel;
}


/* rel_new 를 통해 생성된 인스턴스를 이용해 table의 열기를 수행한다 */
int rel_open(Relation *rel)
{
	
	rel->io = io_fopen(rel->db_file);

	/* 파일 읽기 오류 */
	if(rel->io==NULL)
	{
#ifdef _DEBUG
		fprintf(stderr,"db file open error\n");
#endif
		return -1;
	}

	/* 존재 하는 파일을 그냥 열지 못한다면 */
	if(iostat != IO_OK)
	{
#ifdef _DEBUG
		fprintf(stderr,"db file don't exists\n");
#endif
		return -1;
	}
		
	/* DB 헤더 및 속성 리스트 읽기 */
	if( rel_get_header(rel)!=0 || rel_get_dic(rel) !=0)
	{
#ifdef _DEBUG
		fprintf(stderr,"dic or header read eror\n");
#endif
		return -1; /* fail */
	}

	rel->szrec = rel_get_szrec(rel);

	return 0;		
}


/* rel_new 를 통해 생성된 인스턴스를 이용해 table의 생성을 수행한다 */ 
int rel_creat(Relation *rel)
{
/* io 널 문제 해결 */
	rel->io = io_fopen(rel->db_file);

	if(rel_first_attr(rel) == NULL)
		return -1;

	if(rel_set_header(rel)!=0)
		return -2;
	
	if(rel_set_dic(rel)!=0)
		return -3;

	return 0;
}
		
		
/* Relation 헤더를 디스크로 부터 읽어 값을 채워 넣는다 */
int rel_get_header(Relation *rel)
{
	RelationHeader relheader;
	
	io_fread(rel->io,&relheader,sizeof(RelationHeader));

	if(!(relheader.identity[0] == TINY_FM_IDENTITY[0] &&
		relheader.identity[1] == TINY_FM_IDENTITY[1]))
		return -1;

	rel->rows = relheader.rows;
/*	rel->off_idx = relheader.off_idx; */
	rel->off_data = relheader.off_data;
		
	/* success */
	return 0;
}


/* Relation 헤더의 내용을 디스크로 저장한다 */
int rel_set_header(Relation *rel)
{
	RelationHeader relheader;

	/* FIXME */
	memcpy(&relheader.identity,TINY_FM_IDENTITY,2);
	relheader.version = TINY_FM_VERSION;
	relheader.rows = 0;
/*	relheader.off_idx = sizeof(RelationHeader); */
/*	relheader.off_data = relheader.off_idx + 0; */
	
	io_fseek(rel->io,0);
	io_fwrite(rel->io,&relheader,sizeof(RelationHeader));	

	/* success */
	return 0;
}


/* 사전 파일(*.dic)을 읽어온다 */
int rel_get_dic(Relation *rel)
{
	IO *io;
	DicHeader header;
	AttrEntry attrentry;
	AttrList *tmpattr;
	int count;

	io = io_fopen(rel->dic_file);

	if(iostat != IO_OK)
		return -1;

	io_fseek(io,0);
	io_fread(io,&header,sizeof(DicHeader));
	io_fseek(io,header.off_attr);

	count = rel->count_of_attr = header.count_of_attr;

	while(count > 0)
	{
		io_fread(io,&attrentry,sizeof(AttrEntry));
		tmpattr = rel_new_attr(attrentry.attrname,
				attrentry.domain,attrentry.length);
		rel_insert_attr(rel,tmpattr);
		/* todo - rel->entry 에 추가하는 루틴 */
	
		count--;
	}

	io_fclose(io);

	return 0;
}


/* 사전 정보를 파일에 저장한다 */
int rel_set_dic(Relation *rel)
{
	IO *io;
	DicHeader header;
	AttrEntry attrentry;
	AttrList *tmpattr;


	io = io_fopen(rel->dic_file);

	memset(&header,0,sizeof(DicHeader));
	header.count_of_attr = rel->count_of_attr;
	header.off_attr = sizeof(DicHeader);
	
	/* header 쓸 위치로 이동 */
	io_fseek(io,0);
	io_fwrite(io,&header,sizeof(DicHeader)); /* 헤더 저장 */
	
	/* attr 이 저장된 offset 으로 seek */
	io_fseek(io,header.off_attr);

	/* attr entry list 저장 */
	for(tmpattr = rel_first_attr(rel);tmpattr != NULL;tmpattr = rel_next_attr(tmpattr))
	{
		memset(&attrentry,0,sizeof(AttrEntry));
		
		u_strcpy(attrentry.attrname,tmpattr->attrname);
		attrentry.domain = tmpattr->domain;
		attrentry.length = tmpattr->length;

		io_fwrite(io,&attrentry,sizeof(AttrEntry));
	}
	
	io_fclose(io);

	/* success */
	return 0;
}


/* Relation 인스턴스 소멸자 */
void rel_destroy(Relation *rel)
{
	rel_set_header(rel);

	free(rel->relname);
	free(rel->db_file);
	free(rel->dic_file);
	memset(rel,0,sizeof(Relation));
	free(rel);
}

/* 레코드를 하나 삽입한다 */
int rel_insert_record(Relation *rel,uint32 idx,void *data)	
{
	/* 저장 위치로 이동 */
	io_fseek(rel->io,rel->off_data + rel->szrec * idx);
		
	io_fwrite(rel->io,data,rel->szrec);

	/* success */
	return 0;
}

/* 디스크로부터 레코드를 읽어온다 */
int rel_read_record(Relation *rel,uint32 idx,void *data)	
{
	/* 저장 위치로 이동 */
	io_fseek(rel->io,rel->off_data + rel->szrec * idx);
		
	io_fread(rel->io,data,rel->szrec);

	/* success */
	return 0;
}


/* Tuple 구조체의 속성 요소(Element)의 생성자 */
Element* rel_new_element(void)
{
	Element *elmt = (Element*)u_malloc(sizeof(Element));

	elmt->attr_id = -1;
	elmt->vattr = (Attribute*)u_malloc(sizeof(Element));
	elmt->next = (Element*)NULL;
	
	return elmt;
}

/* Tuple 인스턴스의 속성요소(Element)의 소멸자 */
void rel_destroy_element(Element *elmt)
{
	memset((void*)elmt,0,sizeof(Element));
	u_free(elmt->vattr);
	u_free(elmt);
}


/* Tuple 인스턴스의 생성자 */
Tuple* rel_new_tuple(void)
{
	Tuple *tpl = (Tuple*)u_malloc(sizeof(Tuple));

	tpl->count_of_attr = 0;
	tpl->attrlst = NULL;
	
	/* success */
	return tpl;
}

/* Tuple 인스턴스의 소멸자 */
void rel_destroy_tuple(Tuple *tpl)
{
	Element *tmpelmt;
	Element *remelmt;
	
	tmpelmt = tpl->attrlst;
	
	while(tmpelmt != NULL)
	{
		remelmt = tmpelmt;
		tmpelmt = tmpelmt->next;
		free(remelmt);

		tpl->count_of_attr--;
	}

#ifdef _DEBUG
	fprintf(stderr,"count of left element : %d\n",tpl->count_of_attr);
#endif
}


/* 각 함수 포인터(tpfunc)에 의존해 data값을 Element 에 복사하여
   연결리스트를 생성해 간다  */
int rel_pack_tuple(
		Tuple *tpl,
		Id id, // pack 하는 속성(Attribute 의 ID)
		void (*tpfunc)(Attribute *vattr,void *data), 
		/* Domain type 별로 데이터 처리등을 위한 함수 포인터 */
		void *data)
{
	Element *elmt;
	Element *tmpelmt;

	elmt = (Element*)rel_new_element();

	/* set of new element */ 
	elmt->attr_id = id;
	tpfunc(elmt->vattr,data);
	elmt->next = NULL;

	/* pack 한 element 가 하나도 없을때 */
	if(tpl->attrlst==NULL)
	{
		tpl->attrlst = elmt;
		
	} else { /* 이미 pack 한 element 가 있을 경우 */ 
		
		for(tmpelmt = tpl->attrlst;tmpelmt->next != NULL;
				tmpelmt = tmpelmt->next);

		tmpelmt->next = elmt;
	}
	
	tpl->count_of_attr++;

	/* success */
	return 0;
}


/* 각 도메인 별로 데이터를 공용체에 넣기 위한 콜백 함수들 */
void rel_tpfunc_num(Attribute *vattr, void *data)
{
	memcpy(&vattr->vNum,data,szNumeric);
}

void rel_tpfunc_int(Attribute *vattr, void *data)
{
	memcpy(&vattr->vInt,data,szInt);
}

void rel_tpfunc_char(Attribute *vattr, void *data)
{
	strncpy(vattr->vChar,(char*)data,16);
}


/* packed 한 데이터를 파일에 레코드로 쓰기 위해 raw 데이터로 만든다
   반환되는 (void *) 의 포인터에는 rel->szrec 크기 만큼의 데이터가 들어있다 
 
   FIX ME : 현재 DDL 문에서 속성(Attribute) 를 생성할때 입력하는 각 크기를
		   처리하는 루틴의 추가가 요구된다 */
void* rel_export_raw(Relation *rel,Tuple *tpl)
{
	char *rec=NULL,*write_ptr=NULL;
	Element *tmpelmt = NULL;
	AttrList *attr = NULL;

	rec = (char*)u_malloc(rel->szrec);
	write_ptr = rec;

	tmpelmt = tpl->attrlst;
	
	while(tmpelmt != NULL)
	{

		attr = rel_get_attrobj_withid(rel,tmpelmt->attr_id);
		
		switch(attr->domain)
		{
			case dChar :
				memcpy((void*)write_ptr,(void*)tmpelmt->vattr->vChar,szChar);
				write_ptr+=szChar;							 
				break;

			case dVarchar : /* FIX ME */
				memcpy((void*)write_ptr,(void*)tmpelmt->vattr->vChar,szChar);
				write_ptr+=szVarchar;
				break;

			case dInt : 
				memcpy((void*)write_ptr,(void*)tmpelmt->vattr->vChar,szInt);
				write_ptr+=szInt;
				break;
				
			case dNumeric : 
				memcpy((void*)write_ptr,(void*)tmpelmt->vattr->vChar,szNumeric);
				write_ptr+=szNumeric;
				break;
				
			case dBoolean : break;
			case dClob : break;
			case dBlob : break;
			case dTimestamp : break;
		}

		tmpelmt = tmpelmt->next;
	
	}

	return rec;
}


/* 디스크로 부터 읽은 raw 데이터를 Tuple 구조체로 바꾼다 */
Tuple* rel_export_tuple(Relation *rel,void *data)
{
	Tuple *tpl = NULL;
	AttrList *tmpattr;
	char *read_ptr = data;

	tpl = rel_new_tuple();

	tmpattr = rel_first_attr(rel);

	while(tmpattr != NULL)
	{
		switch(tmpattr->domain)
		{
			case dChar :
				rel_pack_tuple(tpl,tmpattr->id,TPFUNC_CHAR,(void*)read_ptr);
				read_ptr += szChar;
				break;
			
			case dVarchar : 
				rel_pack_tuple(tpl,tmpattr->id,TPFUNC_CHAR,(void*)read_ptr);
				read_ptr += szVarchar;
				break;

			case dInt : 
				rel_pack_tuple(tpl,tmpattr->id,TPFUNC_INT,(void*)read_ptr);
				read_ptr += szInt;
				break;
				
			case dNumeric : 
				rel_pack_tuple(tpl,tmpattr->id,TPFUNC_NUM,(void*)read_ptr);
				read_ptr += szNumeric;
				break;
				
			case dBoolean : 
				read_ptr += szBoolean;
				break;
				
			case dClob : break;
			case dBlob : break;
			case dTimestamp : break;
		}

		tmpattr = tmpattr->next;
	}

	return tpl;
}


/* 테이블 스키마 생성시 속성(AttrList) 생성자 */
AttrList* rel_new_attr(const char *attrname,Domain domain,uint32 length)
{
	AttrList *new;

	if(strlen(attrname) >= MAX_SZ_NAME)
		return (AttrList*)NULL;

	new = (AttrList*)u_malloc(sizeof(AttrList));

	new->id = -1;
	new->attrname = u_strdup(attrname);
	new->domain = domain;
	new->length = length;
	new->next = (AttrList*)NULL;

	return new;
}


/* 테이블 스키마 생성시 속성 삽입 */
int rel_insert_attr(Relation *rel,AttrList *attr)
{
	AttrList *tmp = NULL;
	AttrList *last = NULL;

	if(rel->attrlist == NULL)
	{
		attr->id=0;
		rel->attrlist = attr;
		
	}else{

		for(tmp = rel_first_attr(rel);tmp!=NULL;tmp = rel_next_attr(tmp))
		{
			if(u_strcmp(tmp->attrname,attr->attrname)==0)
			{
				return -1; /* existed name */
			}
			last = tmp;
		}
		attr->id = last->id+1;
		last->next = attr;
	}

	rel->count_of_attr++;

	/* success */
	return 0;
}


/* 레코드의 사이즈를 구하는 함수 */
int rel_get_szrec(Relation *rel)
{
	AttrList *attr;
	uint32 size=0;

	for(attr = rel_first_attr(rel);attr != NULL;attr = rel_next_attr(attr))
	{
		switch(attr->domain)
		{
			case dChar: size += (szChar * attr->length); break;
			case dVarchar: size += (szVarchar * attr->length); break;
			case dInt: size += (szInt * attr->length); break;
			case dNumeric : size += (szNumeric * attr->length); break;
			case dBoolean : size += (szBoolean * attr->length); break;
			case dClob : size += (szClob * attr->length); break;
			case dBlob : size += (szBlob * attr->length); break;
			case dTimestamp : size += (szTimestamp * attr->length); break;
			default : break;
		}

	}

	return size;
}


#ifdef _DEBUG
/* AttrList 목록 출력 */
void rel_printlist_attr(Relation *rel)
{
	AttrList *attr;

	for(attr = rel_first_attr(rel);attr != NULL;attr =  rel_next_attr(attr))
		fprintf(stderr,"Attr id: %d, name : %s, domain : %d, length: %d\n"
				,attr->id,attr->attrname,attr->domain,attr->length);

}


/* Tuple 목록 출력 */
void rel_printlist_tuple(Tuple *tpl)
{
	Element *tmpelmt = tpl->attrlst;
	
	while(tmpelmt != NULL)
	{
		printf("Id: %d\tdata: %d\t%d\t%s\n",tmpelmt->attr_id,
				tmpelmt->vattr->vNum,
				tmpelmt->vattr->vInt,
				tmpelmt->vattr->vChar);

		tmpelmt = tmpelmt->next;
	}
}
#endif


/* 속성 이름으로 id 를 구하는 함수 */
int rel_get_attrid_withname(Relation *rel,const char *name)
{
	AttrList *attr;

	for(attr = rel_first_attr(rel);attr != NULL;attr = rel_next_attr(attr))
		if(u_strcmp(attr->attrname,name)==0)
			return attr->id;

	return -1;
}


/* 속성 이름으로 AttrList 인스턴스를 구하는 함수 */
AttrList* rel_get_attrobj_withname(Relation *rel,const char *name)
{
	AttrList *attr;

	for(attr = rel_first_attr(rel);attr != NULL;attr = rel_next_attr(attr))
		if(u_strcmp(attr->attrname,name)==0)
			return attr;

	return (AttrList*)NULL;
}


/* 속성 ID로 AttrList 인스턴스를 구하는 함수 */
AttrList* rel_get_attrobj_withid(Relation *rel, uint32 id)
{
	AttrList *attr;

	for(attr = rel_first_attr(rel);attr != NULL;attr = rel_next_attr(attr))
		if(id == attr->id)
			return attr;

	return (AttrList*)NULL;
}
