// $Id: relation.h,v 1.1.1.1 2004/11/13 23:28:02 advanced Exp $
//
// B*Tree Header

#ifndef TINYSQL_RELATION_H
#define TINYSQL_RELATION_H

#include<type.h>

#include<btree/io.h>

#define FEXT_DB ".db"
#define FEXT_DIC ".dic"

#define MAX_SZ_NAME 32

#define dChar 0x01
#define dVarchar 0x02
#define dInt 0x03
#define dNumeric 0x04
#define dBoolean 0x05
#define dClob 0x06
#define dBlob 0x07
#define dTimestamp 0x08

#define szChar 16
#define szVarchar 16
#define szInt 4
#define szNumeric 4
#define szBoolean 1
#define szClob 16
#define szBlob 16
#define szTimestamp 4 

typedef uint32 Domain;
typedef uint32 Id;

typedef struct AttrList_ {
	Id id;
	char *attrname;
	Domain domain;
	uint32 length;
	
	struct AttrList_ *next;
} AttrList;
	
typedef struct Relation_ {
	char *relname;
	char *db_file; // db 파일 이름
	char *dic_file; // 사전 파일 이름

	IO *io;

	uint32 rows; // 데이터의 갯수
	uint32 szrec; // record size

/*	offset off_idx; */ /* Tree 영역의 시작 오프셋 */
	offset off_data; // data영역 시작 오프셋
	
	AttrList *attrlist;
	uint32 count_of_attr;
} Relation;

#pragma pack(1)
typedef struct DicHeader_ {
/*	unsigned char identity[2];
	uint16 version;*/
	
	uint32 count_of_attr;
	offset off_attr;
} DicHeader;

typedef struct AttrEntry_ {
	char attrname[MAX_SZ_NAME];
	Domain domain;
	uint32 length;
} AttrEntry;

typedef union Attribute_ {
	uint32 vNum;
	uint32 vInt;
	char vChar[16];
} Attribute;
#pragma pack()

typedef struct Element_ {
	Id attr_id;
	Attribute *vattr;
	struct Element_ *next;
} Element;

typedef struct Tuple_ {
	uint32 count_of_attr;
	Element *attrlst;
} Tuple;


// function list
Relation* rel_new(const char *relname);
void rel_destroy(Relation *tree);
int rel_creat(Relation *rel);
int rel_open(Relation *rel);

int rel_get_dic(Relation *rel);
int rel_set_dic(Relation *rel);

int rel_get_header(Relation *rel);
int rel_set_header(Relation *rel);

int rel_get_szrec(Relation *rel);

int rel_insert_record(Relation *rel,uint32 idx,void *data);
int rel_read_record(Relation *rel,uint32 idx,void *data);
int rel_remove_record(Relation *rel,uint32 idx);


AttrList* rel_new_attr(const char *attrname,Domain domain,uint32 length);
int rel_insert_attr(Relation *rel,AttrList *attr);

int rel_get_attrid_withname(Relation *rel,const char *name);
AttrList* rel_get_attrobj_withid(Relation *rel, uint32 id);
AttrList* rel_get_attrobj_withname(Relation *rel,const char *name);

Tuple* rel_new_tuple(void);
int rel_pack_tuple(Tuple *tpl,
					Id id,
					void (*tpfunc)(Attribute *vattr,void *data),
				    void *data);
void rel_destroy_tuple(Tuple *tpl);


#ifdef _DEBUG /* For Debug */
void rel_printlist_attr(Relation *rel);
void rel_printlist_tuple(Tuple *tpl);
#endif

#define rel_first_attr(rel) (rel)->attrlist
#define rel_next_attr(attr) (attr)->next

void rel_tpfunc_num(Attribute *vattr, void *data);
void rel_tpfunc_int(Attribute *vattr, void *data);
void rel_tpfunc_char(Attribute *vattr, void *data);
void* rel_export_raw(Relation *rel,Tuple *tpl);
Tuple* rel_export_tuple(Relation *rel,void *data);

#define TPFUNC_NUM rel_tpfunc_num
#define TPFUNC_INT rel_tpfunc_int
#define TPFUNC_CHAR rel_tpfunc_char
#endif
