// $Id: readline_1.c,v 1.1.1.1 2004/11/13 23:28:02 advanced Exp $
// A Short Completion Example (From GNU Readline Manual)
#include<stdio.h>
#include<sys/types.h>
#include<sys/file.h>
#include<sys/stat.h>
#include<sys/errno.h>
#include<string.h>

#include<readline/readline.h>
#include<readline/history.h>

#include<utils.h>

int com_list __P((char *));
int com_view __P((char *));
int com_rename __P((char *));
int com_stat __P((char *));
int com_pwd __P((char *));
int com_delete __P((char *));
int com_help __P((char *));
int com_cd __P((char *));
int com_quit __P((char *));


typedef struct {
	char *name;
	rl_icpfunc_t *func;
	char *doc;
} COMMAND ;

COMMAND commands[] = {
	{ "cd", com_cd, "Change to directory DIR" },
	{ "delete", com_delete, "Delete FILE" },
	{ "help", com_help, "Display this text" },
	{ "?", com_help, "Synonym for `help'" },
	{ "list", com_list, "List files in DIR" },
	{ "ls", com_list, "Synonym for `list'" },
	{ "pwd", com_pwd, "Print the current working directory" },
	{ "quit", com_quit, "Quit using Fileman" },
	{ "rename", com_rename, "Rename FILE to NEWNAME" },
	{ "stat", com_stat, "Print out statistics on FILE" },
	{ "view", com_view, "View the contents of FILE" },
	{ (char *)NULL, (rl_icpfunc_t *)NULL, (char *)NULL }
};

char *stripwhite();
COMMAND *find_command();

char *progname;

int done;


char *dupstr(char *s)
{
	char *r;

	r = xmalloc(strlen(s)+1);
	strcpy(r,s);
	return r;
}

int main(int argc,char *argv[])
{
	char *line, *s;

	progname = argv[0];

	initialize_readline();

	for(;done==0;)
	{
		line = readline("FileMan: ");

		if(!line)
			break;

		s = stripwhite(line);

		if(*s)
		{
			add_history(s);
			execute_line(s);
		}

		free(line);
	}

	exit(0);
}

int execute_line(char *line)
{
	register int i;
	COMMAND *command;
	char *word;

	i = 0;
	while(line[i] && whitespace(line[i]))
		i++;
	word = line + i;

	while(line[i] && !whitespace(line[i]))
		i++;

	if(line[i])
		line[i++] = '\0';

	command = find_command(word);

	if(!command)
	{
		fprintf(stderr,"%s: No such command for FileMan.\n",word);
		return -1;
	}

	while(whitespace(line[i]))
		i++;

	word = line + i;

	return ((*(command->func))(word));
}

COMMAND *find_command(char *name)
{
	register int i;

	for(i=0;commands[i].name;i++)
		if(strcmp(name,commands[i].name) == 0)
			return (&commands[i]);

	return ((COMMAND *)NULL);
}

char *stripwhite(char *string)
{
	register char *s,*t;

	for(s = string;whitespace(*s);s++)
		;

	if(*s == 0)
		return (s);

	t = s+strlen(s) -1;
	while(t>s && whitespace(*t))
		t--;
	*++t = '\0';

	return s;
}

char *command_generator __P((const char *,int ));
char **fileman_completion __P((const char *,int,int));

initialize_readline()
{

	rl_readline_name="FileMan";
	rl_attempted_completion_function = fileman_completion;
	rl_bind_key('\t',rl_insert);
//	rl_attempted_completion_function = abc;
//	rl_filename_quoting_function = rl_completion_entry_function;

}

char **fileman_completion(const char *text,int start,int end)
{
	char **matches;

	matches = (char **)NULL;

	if(start ==0)
		matches = rl_completion_matches(text,command_generator);

	return (matches);
}

char *command_generator(const char *text,int state)
{
	static int list_index,len;
	char *name;

	if(!state)
	{
		list_index = 0;
		len = strlen(text);
	}

	while(name = commands[list_index].name)
	{
		list_index++;

		if(strncmp(name,text,len) == 0)
			return (dupstr(name));
	}
		return ((char *)NULL);
}
		
static char syscom[1024];

com_list(char *arg)
{
	if(!arg)
		arg="";

	sprintf(syscom,"ls -FClg %s",arg);
	return (system(syscom));
}

com_view(char *arg)
{
	if(!valid_argument("view",arg))
		return 1;

	sprintf(syscom,"more %s",arg);
	return (system(syscom));
}

com_rename(char *arg)
{
	too_dangerous("rename");
	return 1;
}

com_stat(char *arg)
{
	struct stat finfo;

	if(!valid_argument("stat",arg))
		return 1;

	if(stat(arg,&finfo)==-1)
	{
		perror(arg);
		return 1;
	}

	printf("Statistics for `%s':\n",arg);

	printf("%s has %d link %s, and is %d byte %s in length.\n",arg,
			finfo.st_nlink,
			(finfo.st_nlink==1)?"":"s",
			finfo.st_size,
			(finfo.st_size==1)?"":"s");

	printf("inode Last Change at: %s",ctime(&finfo.st_ctime));
	printf("	Last access at: %s",ctime(&finfo.st_atime));
	printf("	Last modified at: %s",ctime(&finfo.st_mtime));

	return 0;
}

int com_delete(char *arg)
{
	too_dangerous("delete");
	return 1;
}

int com_help(char *arg)
{
	register int i;
	int printed = 0;

	for(i = 0;commands[i].name;i++)
	{
		if(!*arg||(strcmp(arg,commands[i].name)==0))
		{
			printf("%s\t\t%s.\n",commands[i].name,commands[i].doc);
			printed++;
		}
	}

	if(!printed)
	{
		printf("No commands match `%s'. Posibilities are:\n",arg);

		for(i=0;commands[i].name;i++)
		{
			if(printed==6)
			{
				printed=0;
				printf("\n");
			}

			printf("%s\t",commands[i].name);
			printed++;
		}

		if(printed)
			printf("\n");

	}
	return 0;
}

com_cd(char *arg)
{
	if(chdir(arg)==-1)
	{
		perror(arg);
		return 1;
	}

	com_pwd("");

	return 0;
}

com_pwd(char *ignore)
{
	char dir[1024],*s;

	s = getcwd(dir,sizeof(dir)-1);

	if(s==0)
	{
		printf("Error getting pwd: %s\n",dir);
		return 1;
	}

	printf("Current directory is %s\n",dir);
	return 0;
}

com_quit(char *arg)
{
	done = 1;
	return 0;
}

too_dangerous(char *caller)
{
	fprintf(stderr,"%s: Too dangerous for me to distribute.\n",caller);
	fprintf(stderr,"Write it yourself.\n");
}


int valid_argument(char *caller,char *arg)
{
	if (!arg || !*arg)
	{
		fprintf(stderr,"%s: Argument required.\n",caller);
		return 0;
	}

	return 1;
}	
