// $Id: sqlparser.yy,v 1.1.1.1 2005/02/28 05:37:15 advanced Exp $
// Reference - http://acdbt.org/wiki/wiki.php?tinySQL/SQL

%{
#include<iostream>
#include<sstream>
#include<FlexLexer.h>

#include"sqlproc.h"

#define yylex() (lexer->yylex())

using namespace std;

static FlexLexer *lexer=NULL;

void yyerror(char *str);
void parse(stringstream *ss);


%}

%union 
{
	long ival;
	double fval;
	char chr;
	char *str;
	const char *keyword;
	struct Node *node;
	struct DataType *datatype;
	struct TableElmt *tbelmt;
}

%type <node> dml_statement
%type <node> ddl_statement
%type <str> qualified_id table_name
%type <datatype> data_type
%type <tbelmt> table_element
%type <node> table_element_list

// SQL92 reserved keyword
%token <keyword> ABSOLUTE  ACTION  ADD  ALL
	ALLOCATE  ALTER  AND
	ANY  ARE
	AS  ASC
	ASSERTION  AT
	AUTHORIZATION  AVG
	BEGIN_  BETWEEN  BIT  BIT_LENGTH
	BOTH  BY
	CASCADE  CASCADED  CASE  CAST
	CATALOG
	CHAR  CHARACTER  CHAR_LENGTH
	CHARACTER_LENGTH  CHECK  CLOSE  COALESCE
	COLLATE  COLLATION
	COLUMN  COMMIT
	CONNECT
	CONNECTION  CONSTRAINT
	CONSTRAINTS  CONTINUE
	CONVERT  CORRESPONDING  COUNT  CREATE  CROSS
	CURRENT
	CURRENT_DATE  CURRENT_TIME
	CURRENT_TIMESTAMP  CURRENT_USER  CURSOR
	DATE  DAY  DEALLOCATE  DEC
	DECIMAL  DECLARE  DEFAULT  DEFERRABLE
	DEFERRED  DELETE  DESC  DESCRIBE  DESCRIPTOR
	DIAGNOSTICS
	DISCONNECT  DISTINCT  DOUBLE  DROP
	ELSE  END  END_EXEC  ESCAPE
	EXCEPT  EXCEPTION
	EXEC  EXECUTE  EXISTS
	EXTERNAL  EXTRACT
	FALSE  FETCH  FIRST  FLOAT  FOR
	FOREIGN  FOUND  FROM  FULL
	GET  GLOBAL  GO  GOTO
	GRANT  GROUP
	HAVING  HOUR
	IDENTITY  IMMEDIATE  IN  INDICATOR
	INITIALLY  INNER  INPUT
	INSENSITIVE  INSERT  INT  INTEGER  INTERSECT
	INTERVAL  INTO  IS
	ISOLATION
	JOIN
	KEY
	LANGUAGE  LAST  LEADING  LEFT
	LEVEL  LIKE  LOCAL  LOWER
	MATCH  MAX  MIN  MINUTE  MODULE
	MONTH
	NAMES  NATIONAL  NATURAL  NCHAR  NEXT  NO
	NOT  _NULL_
	NULLIF  NUMERIC
	OCTET_LENGTH  OF
	ON  ONLY  OPEN  OPTION  OR
	ORDER  OUTER
	OUTPUT  OVERLAPS
	PAD  PARTIAL  POSITION  PRECISION  PREPARE
	PRESERVE  PRIMARY
	PRIOR  PRIVILEGES  PROCEDURE  PUBLIC
	READ  REAL  REFERENCES  RELATIVE  RESTRICT
	REVOKE  RIGHT
	ROLLBACK  ROWS
	SCHEMA  SCROLL  SECOND  SECTION
	SELECT
	SESSION  SESSION_USER  SET
	SIZE  SMALLINT  SOME  SPACE  SQL  SQLCODE
	SQLERROR  SQLSTATE
	SUBSTRING  SUM  SYSTEM_USER
	TABLE  TEMPORARY
	THEN  TIME  TIMESTAMP
	TIMEZONE_HOUR  TIMEZONE_MINUTE
	TO  TRAILING  TRANSACTION
	TRANSLATE  TRANSLATION  TRIM  TRUE
	UNION  UNIQUE  UNKNOWN  UPDATE  UPPER  USAGE
	USER  USING
	VALUE  VALUES  VARCHAR  VARYING  VIEW
	WHEN  WHENEVER  WHERE  WITH  WORK  WRITE
	YEAR
	ZONE

// internal token
%token <keyword>ASTERISK
	CHARACTER_STRING_LITERAL COMMA
	CONCATENATION_OPERATOR
	EQUALS_OPERATOR
	DELIMITED_IDENTIFIER 
	GREATER_THAN_OPERATOR GREATER_THAN_OR_EQUALS_OPERATOR
	LESS_THAN_OPERATOR LESS_THAN_OR_EQUALS_OPERATOR
	MINUS_SIGN
	NOT_EQUALS_OPERATOR
	PERIOD PLUS_SIGN
	UNDERSCORE
	QUOTE QUOTE_SYMBOL
	SEMICLON SOLIDUS

%token <str> IDENTIFIER
%token <ival> DIGIT

%left '(' ')'


%start sql
%%

// START SYMBOL

sql :
	sql_statement SEMICLON {
#ifdef DEBUG
	sql_ok();
#endif
	}
	;
sql_statement:
	// Data manipulation language
	dml_statement {
		cout << $1->stmt <<endl;
	} 
	| // Data definition language
	ddl_statement {
		cout << $1->stmt <<endl;
	};

///////////////////////////////////////////////////////////////////////////////
dml_statement:
	// select statement
	SELECT select_list FROM table_name { 
		Node *node = new_node(_SELECT); 
		$$ = node;
	} |
	SELECT select_list FROM table_name where_clause {
		Node *node = new_node(_SELECT);
		$$ = node;
	} |
	// update statement
	UPDATE table_name {
		Node *node = new_node(_UPDATE);
		$$ = node;
	};

select_list:
	ASTERISK |
	select_sublist_list;

select_sublist_list :
	select_sublist_list COMMA select_sublist |
	select_sublist ;

select_sublist:
	IDENTIFIER PERIOD ASTERISK |
	IDENTIFIER PERIOD IDENTIFIER PERIOD ASTERISK |
	derived_column ;

derived_column:
	IDENTIFIER |
	IDENTIFIER as_clause;

as_clause:
	AS IDENTIFIER;

where_clause:
	WHERE ;


///////////////////////////////////////////////////////////////////////////////
// data definition state
ddl_statement:
	// table definition statement
	CREATE TABLE table_name '(' table_element_list ')' {
		cout << "inserting table:" <<  $3 << endl;

		vector<TableElmt*>::const_iterator pos;

		for(pos = $5->tables.begin();pos != $5->tables.end();
			++pos)
		{
			cout << "created attr:" << (*pos)->tbname << endl;
		}

		$$ = $5;
	}; 



table_name :
	qualified_id;

table_element_list:
	table_element_list COMMA table_element {
		$1->tables.push_back($3);
		$$ = $1;
	}
	|
	table_element {
		$$ = new_node(_TABLE,$1);
	};

table_element:
	table_name data_type {
		$$ = new_table_elmt($1,$2);
	};

data_type:
	INT {
		$$=new_data_type(0);
	}
	|
	INT '(' DIGIT ')' {
		$$=new_data_type(0,$3);
	}
	|
	CHAR {
		$$=new_data_type(1);
	}
	|
	CHAR '(' DIGIT ')' {
		$$=new_data_type(1,$3);
	};
///////////////////////////////////////////////////////////////////////////////	
/*qualifier :
	qualified_id dotted_name |
	qualified_id
	;

dotted_name:
	PERIOD qualified_id;*/

qualified_id:
	IDENTIFIER {
		$$ = strndup($1,strlen($1));
	};
	
%%

void yyerror(char *str)
{
	cout << str << endl;
}

void parse(stringstream *syyin)
{
	lexer = new yyFlexLexer(syyin);

	yyparse();
}
